from datetime import datetime, timezone
from typing import NoReturn, TypeVar
import requests as r
from requests.exceptions import HTTPError
import logging
import sys
from dacite import Config, from_dict

T = TypeVar("T")


class Utils:
    def __init__(self, url, api_key):
        self.log = logging.getLogger("immich-tools")
        self.url = url
        self.api_key = api_key

    def send_delete(self, path: str, data: dict):
        try:
            response = r.delete(self.url + path, json=data, headers={"x-api-key": self.api_key})
            response.raise_for_status()
        except r.exceptions.HTTPError as err:
            self.__log_and_exit(err, path)
        return response

    def datetime_to_str(self, datetime: datetime) -> str:
        dt = datetime.now(timezone.utc)
        formatted = dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
        return formatted

    def send_multipart(self, path: str, files, data):
        try:
            response = r.post(self.url + path, files=files, headers={"x-api-key": self.api_key}, data=data)
            response.raise_for_status()
        except r.exceptions.HTTPError as err:
            self.__log_and_exit(err, path)
        return response

    def send_put(self, path: str, data: dict = {}):
        try:
            response = r.put(self.url + path, json=data, headers={"x-api-key": self.api_key})
            response.raise_for_status()
        except r.exceptions.HTTPError as err:
            self.__log_and_exit(err, path)
        return response

    def send_get(self, path: str) -> r.Response:
        try:
            response = r.get(self.url + path, headers={"x-api-key": self.api_key})
            response.raise_for_status()
        except HTTPError as err:
            self.__log_and_exit(err, path)
        return response

    def send_post(self, path: str, data: dict = {}):
        try:
            response = r.post(self.url + path, json=data, headers={"x-api-key": self.api_key})
            response.raise_for_status()
        except r.exceptions.HTTPError as err:
            self.__log_and_exit(err, path)
        return response

    def __snake_to_camel(self, name: str) -> str:
        components = name.split("_")
        s = components[0] + "".join(x.title() for x in components[1:])
        return s

    def from_dict(self, data_class: type[T], data: dict) -> T:
        return from_dict(
            data_class=data_class,
            data=data,
            config=Config(
                type_hooks={datetime: lambda s: datetime.fromisoformat(s.replace("Z", "+00:00"))},
                convert_key=self.__snake_to_camel,
            ),
        )

    def __log_and_exit(self, err: HTTPError, path: str) -> NoReturn:
        self.log.error(f"got error for request {self.url + path}: {err}")
        sys.exit(-1)



import click
from immich_tools.immich.client.ImmichClient import ImmichClient
from immich_tools.immich.model.Album import Album
from immich_tools.immich.model.Asset import Asset
import logging
import io
import mimetypes

log = logging.getLogger("immich-tools")
global immich


@click.command()
@click.argument("album_id")
@click.option("-k", "--api-key", required=True, envvar="IMMICH_API_KEY")
@click.option("-u", "--url", required=True, envvar="IMMICH_URL")
def change_owner_photo(album_id, api_key, url):
    """change ownership of all assets in the album to you. Use all as album id to change ownership of all albums"""
    log.debug(f"immich url: {url}")
    global immich
    immich = ImmichClient(url, api_key)
    if album_id == "all":
        albums = immich.get_albums()
        for album in albums:
            __handle_album(immich.get_album(album.id))
    else:
        album = immich.get_album(album_id)
        __handle_album(album)
    
    log.info("success")


def __handle_album(album: Album):
    log.debug(f"found album: {album.album_name}")
    owner_id = __get_owner_of_api_key()
    assets = album.assets
    for asset in assets:
        old_asset_id = asset.id
        old_asset_info = immich.get_asset(asset.id)
        if old_asset_info.owner_id != owner_id:
            log.debug(f"asset: {old_asset_id} is not yours, downloading...")
            asset_content = immich.get_original_asset(old_asset_id)
            live_video_id = asset.live_photo_video_id
            if live_video_id != None:
                log.debug(f"asset have live video related: {live_video_id}, downloading...")
                live_video_asset = immich.get_asset(live_video_id)
                live_video_content = immich.get_asset_content(live_video_id)
                __upload_file(live_video_content, live_video_asset)
            new_asset_id = __upload_file(asset_content, asset)
            log.debug(f"adding new asset {new_asset_id} to album...")
            immich.add_asset_to_album(album.id, new_asset_id)
            log.debug(f"new asset added, now removing old asset from album: {old_asset_id}")
            immich.remove_asset_from_album(album.id, old_asset_id)

def __get_owner_of_api_key():
    return immich.get_user().id


def __upload_file_old(url: str, api_key: str, content: bytes, asset: Asset):
    filename = asset.original_file_name
    mime_type, _ = mimetypes.guess_type(filename)
    files = {"assetData": (filename, io.BytesIO(content), mime_type)}
    data = {
        "deviceAssetId": asset.device_asset_id,
        "deviceId": asset.device_id,
        "fileCreatedAt": asset.file_created_at,
        "fileModifiedAt": asset.file_modified_at,
    }
    log.debug(f"uploading {filename}...")
    r = immich.utils.send_multipart("/api/assets", files, data)
    return r.json()["id"]


def __upload_file(content: bytes, asset: Asset):
    return immich.upload_asset(
        content,
        asset.device_asset_id,
        asset.device_id,
        asset.file_created_at,
        asset.file_modified_at,
        asset.original_file_name,
    ).id



from datetime import datetime

import logging
import mimetypes
import io


from immich_tools.immich.model.Album import Album
from immich_tools.immich.model.Asset import Asset
from immich_tools.immich.model.response.AssetToAlbumResponse import AssetToAlbumResponse
from immich_tools.immich.model.response.AssignTagResponse import AssignTagResponse
from immich_tools.immich.model.Tag import Tag
from immich_tools.immich.model.User import User
from immich_tools.immich.model.response.CreateTagResponse import CreateTagResponse
from immich_tools.immich.model.response.UploadAssetResponse import UploadAssetResponse
from immich_tools.immich.utils.Utils import Utils


class ImmichClient:
    def __init__(self, url, api_key):
        self.log = logging.getLogger("immich-tools")
        self.url = url
        self.api_key = api_key
        self.utils = Utils(url, api_key)

    def get_asset(self, id: str) -> Asset:
        """Returns asset of id \n
        path: /api/asssets/:id

        Parameters:
          id: id of the asset

        Returns:
          asset
        """
        r = self.utils.send_get(f"/api/assets/{id}")
        return self.utils.from_dict(data_class=Asset, data=r.json())

    def get_album(self, id: str) -> Album:
        """Returns album of id \n
        path: /api/albums/:id

        Parameters:
          id: id of the album

        Returns:
          album
        """
        r = self.utils.send_get(f"/api/albums/{id}")
        return self.utils.from_dict(Album, r.json())
      
      
    def get_albums(self) -> list[Album]:
        """Returns album of id \n
        path: /api/albums/:id

        Returns:
          all album
        """
        r = self.utils.send_get(f"/api/albums")
        albums_json = r.json()
        return [self.utils.from_dict(Album, album_json) for album_json in albums_json]

    def get_all_tags(self) -> list[Tag]:
        """Returns all tags \n
        path: /api/tags

        Returns:
          list of tags
        """
        r = self.utils.send_get(f"/api/tags")
        tags_json = r.json()
        return [self.utils.from_dict(Tag, tag_json) for tag_json in tags_json]

    def assign_tag_to_assets(self, tagId: str, assetIds: list[str]) -> list[AssignTagResponse]:
        """assign tags to assets \n
        PUT /api/tags/{tagId}/assets

        Returns:
          Immich Response
        """
        r = self.utils.send_put(f"/api/tags/{tagId}/assets", data={"ids": assetIds})
        responses = r.json()
        return [self.utils.from_dict(AssignTagResponse, response) for response in responses]

    def get_user(self) -> User:
        """Returns owner of the api key \n
        GET /api/users/me

        Returns:
          User
        """
        r = self.utils.send_get(f"/api/users/me")
        return self.utils.from_dict(User, r.json())

    def create_tag(self, name: str, color="", parentId="") -> CreateTagResponse:
        """create tag \n
        POST /api/tags
        """
        data = {"name": name}
        if color and color != "":
            data["color"] = color
        if parentId and parentId != "":
            data["parentId"] = parentId
        r = self.utils.send_post(f"/api/tags", data=data)
        return self.utils.from_dict(CreateTagResponse, r.json())
      
    def upsert_tags(self, names: list[str]) -> list[CreateTagResponse]:
        """create or update tags \n
        PUT /api/tags
        """
        data = {"tags": names}
        r = self.utils.send_put(f"/api/tags", data=data)
        responses = r.json()
        return [self.utils.from_dict(CreateTagResponse, response) for response in responses]

    def get_asset_content(self, assetId: str) -> bytes:
        """Returns owner of the api key \n
        GET /api/assets/{id}/original

        Returns:
          content of asset in bytes
        """
        r = self.utils.send_get(f"/api/assets/{assetId}/original")
        return r.content

    def upload_asset(
        self,
        asset_data: bytes,
        device_asset_id: str,
        device_id: str,
        file_created_at: datetime,
        file_modified_at: datetime,
        filename: str,
    ) -> UploadAssetResponse:
        """upload asset \n
        POST /assets

        Returns:
          UploadAssetResponse
        """
        mime_type, _ = mimetypes.guess_type(filename)
        files = {"assetData": (filename, io.BytesIO(asset_data), mime_type)}
        data = {
            "deviceAssetId": device_asset_id,
            "deviceId": device_id,
            "fileCreatedAt": self.utils.datetime_to_str(file_created_at),
            "fileModifiedAt": self.utils.datetime_to_str(file_modified_at),
        }
        self.log.debug(f"uploading {filename}...")
        r = self.utils.send_multipart("/api/assets", files, data)
        return self.utils.from_dict(UploadAssetResponse, r.json())

    def add_asset_to_album(self, album_id: str, asset_id: str) -> AssetToAlbumResponse:
        """add asset to album \n
        PUT /api/albums/{album_id}/assets

        Returns:
          AddAssetToAlbumResponse
        """
        data = {"ids": [asset_id]}
        r = self.utils.send_put(f"/api/albums/{album_id}/assets", data)
        return self.utils.from_dict(AssetToAlbumResponse, r.json())

    def remove_asset_from_album(self, album_id: str, asset_id: str) -> AssetToAlbumResponse:
        """remove asset from album \n
        DELETE /api/albums/{album_id}/assets

        Returns:
          AssetToAlbumResponse
        """
        data = {"ids": [asset_id]}
        r = self.utils.send_delete(f"/api/albums/{album_id}/assets", data)
        return self.utils.from_dict(AssetToAlbumResponse, r.json())

    def get_original_asset(self, asset_id: str) -> bytes:
        """get original asset \n
        GET /api/assets/{asset_id}/original

        Returns:
          original asset
        """
        r = self.utils.send_get(f"/api/assets/{asset_id}/original")
        return r.content


import click
from immich_tools.immich.client.ImmichClient import ImmichClient, Album
import logging
import sys

log = logging.getLogger("immich-tools")
global immich


@click.command()
@click.argument("album_id")
@click.option("-k", "--api-key", required=True, envvar="IMMICH_API_KEY")
@click.option("-u", "--url", required=True, envvar="IMMICH_URL")
@click.option("-c", "--create-tag", is_flag=True, help="will create missing tags if needed")
def assign_album_tags(album_id, api_key, url, create_tag):
    """looking in album for specified tags and return assets without expected tags. Use all as album id to assign tags in all albums"""
    log.debug(f"immich url: {url}")
    global immich
    immich = ImmichClient(url, api_key)
    if album_id == "all":
        albums = immich.get_albums()
        for album in albums:
            __handle_album(immich.get_album(album.id))
    else:
        album = immich.get_album(album_id)
        __handle_album(album)
    log.info("success")

def __handle_album(album: Album):
    album_name = album.album_name
    log.debug(f"found album: {album_name}")
    tags = __get_tags_from_discription(album.description)
    asset_ids = list(map(lambda a: a.id, album.assets))
    log.debug("creating tags if necessary")
    create_tag_responses = immich.upsert_tags(tags)
    for response in create_tag_responses:
        log.debug(f"assigning tag: {response.value}")
        immich.assign_tag_to_assets(response.id, asset_ids)

def __get_tags_from_discription(description: str) -> list[str]:
    tags = []
    for line in description.split("\n"):
        if "=" in line and line.split("=")[0] == "tag":
            tag = line.split("=")[1]
            log.debug(f"found tag in description: {tag}")
            tags.append(tag)
            
    return tags

def __get_all_tags_value_id() -> dict:
    all_tags = immich.get_all_tags()
    result = {}
    for tag in all_tags:
        result[tag.value] = tag.id
    return result


def __add_missing_tags(asset_id: str, missing_tags: list[str], all_tags: dict, create_tag: bool):
    for missing_tag in missing_tags:
        log.debug(f"adding tag: {missing_tag} to asset: {asset_id}")
        if missing_tag not in all_tags:
            if create_tag:
                log.debug(f"creating tag: {missing_tag}")
                immich.create_tag(name=missing_tag)
                log.debug("retrying to add tag to asset...")
                __add_missing_tags(asset_id, missing_tags, __get_all_tags_value_id(), create_tag)
                return
            else:
                log.error(f'there is not added tag "{missing_tag}" in immich. Add tag then retry.')
                sys.exit(-1)
        id = all_tags[missing_tag]
        immich.assign_tag_to_assets(id, [asset_id])



import click
from immich_tools.immich.client.ImmichClient import ImmichClient
import logging
import sys

log = logging.getLogger("immich-tools")
global immich


@click.command()
@click.argument("album_id")
@click.option("-k", "--api-key", required=True, envvar="IMMICH_API_KEY")
@click.option("-u", "--url", required=True, envvar="IMMICH_URL")
@click.option(
    "-t", "--tag", "expected_tags", required=True, multiple=True, help="expected tags (use multiple -t/--tag)"
)
@click.option("-a", "--add-missing", is_flag=True, help="will add missing tags")
@click.option("-c", "--create-tag", is_flag=True, help="will create missing tags if needed")
def check_album_tags(album_id, api_key, url, expected_tags, add_missing, create_tag):
    """looking in album for specified tags and return assets without expected tags"""
    log.debug(f"immich url: {url}")
    global immich
    immich = ImmichClient(url, api_key)
    album = immich.get_album(album_id)
    album_name = album.album_name
    log.debug(f"found album: {album_name}")
    assets = list(map(lambda a: a.id, album.assets))
    for asset_id in assets:
        asset = immich.get_asset(asset_id)
        tags = asset.tags if asset.tags else []
        tags = list(map(lambda tag: tag.value, tags))
        if not set(expected_tags).issubset(tags):
            missing_tags = list(set(expected_tags) - set(tags))
            print(f"asset: {asset.id}, actual tags: {tags}, missing tags: {missing_tags}")
            if add_missing:
                all_tags = __get_all_tags_value_id()
                __add_missing_tags(asset.id, missing_tags, all_tags, create_tag)
    log.info("success")


def __get_all_tags_value_id() -> dict:
    all_tags = immich.get_all_tags()
    result = {}
    for tag in all_tags:
        result[tag.value] = tag.id
    return result


def __add_missing_tags(asset_id: str, missing_tags: list[str], all_tags: dict, create_tag: bool):
    for missing_tag in missing_tags:
        log.debug(f"adding tag: {missing_tag} to asset: {asset_id}")
        if missing_tag not in all_tags:
            if create_tag:
                log.debug(f"creating tag: {missing_tag}")
                immich.create_tag(name=missing_tag)
                log.debug("retrying to add tag to asset...")
                __add_missing_tags(asset_id, missing_tags, __get_all_tags_value_id(), create_tag)
                return
            else:
                log.error(f'there is not added tag "{missing_tag}" in immich. Add tag then retry.')
                sys.exit(-1)
        id = all_tags[missing_tag]
        immich.assign_tag_to_assets(id, [asset_id])



from dataclasses import dataclass
from datetime import datetime


@dataclass
class User:
  avatar_color: str
  created_at: datetime
  deleted_at: datetime | None
  email: str
  id: str
  is_admin: bool
  name: str
  oauth_id: str
  profile_changed_at: str
  profile_image_path: str
  quota_size_in_bytes: int | None
  quota_usage_in_bytes: int
  should_change_password: bool
  status: str
  storage_label: str
  updated_at: datetime
  


from dataclasses import dataclass
from datetime import datetime

from immich_tools.immich.model.Tag import Tag

@dataclass
class Asset:
    checksum: str
    device_asset_id: str
    device_id: str
    duplicate_id: str | None
    duration: str
    file_created_at: datetime
    file_modified_at: datetime
    has_metadata: bool
    id: str
    is_archived: bool
    is_favorite: bool
    is_offline: bool
    is_trashed: bool
    live_photo_video_id: str | None
    local_date_time: datetime
    original_file_name: str
    original_mime_type: str | None
    original_path: str
    owner_id: str
    thumbhash: str | None
    type: str
    updated_at: str
    visibility: str
    tags: list[Tag] | None


import pytest
from click.testing import CliRunner
from src.immich_tools.cli import main


@pytest.fixture
def runner():
    return CliRunner()


def test_main_help(runner):
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert "This is immich-tools version" in result.output


def test_main_refresh_album_metadata_commands_registered(runner):
    result = runner.invoke(main, ["refresh-album-metadata", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output


def test_main_merge_xmp_command_registered(runner):
    result = runner.invoke(main, ["merge-xmp", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output


def test_main_run_job_command_registered(runner):
    result = runner.invoke(main, ["run-job", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output


def test_main_version_command_registered(runner):
    result = runner.invoke(main, ["version", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output


def test_main_check_album_tags_command_registered(runner):
    result = runner.invoke(main, ["check-album-tags", "--help"])
    assert result.exit_code == 0
    assert "Usage:" in result.output



from setuptools import setup, find_packages
import os

version = {}
with open(os.path.join("src", "immich_tools", "__init__.py")) as f:
    exec(f.read(), version)

setup(
    name="immich-tools",
    version=version["__version__"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=["click", "requests", "PyExifTool", "pytest", "dacite"],
    entry_points={
        "console_scripts": [
            "immich-tools = immich_tools.cli:main",
        ],
    },
)



import click
from .commands import (
    check_album_tags,
    merge_xmp,
    refresh_album_metadata,
    run_job,
    version,
    change_owner_photo,
    assign_album_tags
)
import logging
from datetime import datetime
import immich_tools
import os


def __setup_logging(debug: bool, log_path: str):
    """Configure logging for the CLI."""
    log = logging.getLogger("immich-tools")
    log_level = logging.DEBUG if debug else logging.INFO
    log.setLevel(log_level)
    formatter = logging.Formatter("%(asctime)s - %(levelname)-5s - %(message)s")
    log.handlers.clear()
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    log.addHandler(console_handler)

    if log_path:
        log_directory = log_path or "."
        os.makedirs(log_directory, exist_ok=True)
        log_filename = datetime.now().strftime("log_immich-tools_%d_%m_%Y.log")
        file_path = os.path.join(log_directory, log_filename)
        file_handler = logging.FileHandler(file_path)
        file_handler.setLevel(log_level)
        file_handler.setFormatter(formatter)
        log.addHandler(file_handler)
    return log


@click.group(help=f"""This is immich-tools version {immich_tools.__version__}""")
@click.option("-d", "--debug", is_flag=True, help="Enable debug mode")
@click.option(
    "-l", "--log-path", help="path to directory where logs will be stored, works only if --log-file flag is set to True"
)
@click.pass_context
def main(ctx, debug: bool, log_path: str):
    ctx.ensure_object(dict)
    log = __setup_logging(debug, log_path)
    log.info("Running immich-tools")
    if debug:
        log.debug("Debug mode is ON")


commands = [
    refresh_album_metadata.refresh_album_metadata,
    merge_xmp.merge_xmp,
    run_job.run_job,
    version.version,
    check_album_tags.check_album_tags,
    change_owner_photo.change_owner_photo,
    assign_album_tags.assign_album_tags
]

for command in commands:
    main.add_command(command)



from dataclasses import dataclass
from datetime import datetime
from .Asset import Asset

@dataclass
class Album:
    album_name: str
    album_thumbnail_asset_id: str
    asset_count: int
    assets: list[Asset]
    created_at: datetime
    description: str
    end_date: datetime | None
    has_shared_link: bool
    id: str
    is_activity_enabled: bool
    last_modified_asset_timestamp: str | None
    order: str | None
    owner_id: str
    shared: bool
    start_date: datetime | None
    updated_at: datetime
    